import { writeImplicitNeverMessage } from "../../parse/string/ast.js";
import { throwParseError } from "../../utils/errors.js";
import { listFrom } from "../../utils/generics.js";
import { composeIntersection, composeKeyedIntersection, equality, isDisjoint, isEquality } from "../compose.js";
import { classIntersection } from "./class.js";
import { collapsibleListUnion } from "./collapsibleSet.js";
import { divisorIntersection } from "./divisor.js";
import { flattenObjectKind, objectKindIntersection } from "./objectKind.js";
import { flattenProps, propsIntersection } from "./props.js";
import { flattenRange, rangeIntersection } from "./range.js";
import { getRegex, regexIntersection } from "./regex.js";
const rulesOf = (branch)=>branch.input ?? branch;
export const branchIntersection = (l, r, state)=>{
    const lRules = rulesOf(l);
    const rRules = rulesOf(r);
    const rulesResult = rulesIntersection(lRules, rRules, state);
    if ("morph" in l) {
        if ("morph" in r) {
            if (l.morph === r.morph) {
                return isEquality(rulesResult) || isDisjoint(rulesResult) ? rulesResult : {
                    input: rulesResult,
                    morph: l.morph
                };
            }
            return state.lastOperator === "&" ? throwParseError(writeImplicitNeverMessage(state.path, "of morphs")) : {};
        }
        return isDisjoint(rulesResult) ? rulesResult : {
            input: isEquality(rulesResult) ? l.input : rulesResult,
            morph: l.morph
        };
    }
    if ("morph" in r) {
        return isDisjoint(rulesResult) ? rulesResult : {
            input: isEquality(rulesResult) ? r.input : rulesResult,
            morph: r.morph
        };
    }
    return rulesResult;
};
export const rulesIntersection = (l, r, state)=>"value" in l ? "value" in r ? l.value === r.value ? equality() : state.addDisjoint("value", l.value, r.value) : literalSatisfiesRules(l.value, r, state) ? l : state.addDisjoint("leftAssignability", l, r) : "value" in r ? literalSatisfiesRules(r.value, l, state) ? r : state.addDisjoint("rightAssignability", l, r) : narrowableRulesIntersection(l, r, state);
const narrowIntersection = composeIntersection(collapsibleListUnion);
export const narrowableRulesIntersection = composeKeyedIntersection({
    objectKind: objectKindIntersection,
    divisor: divisorIntersection,
    regex: regexIntersection,
    props: propsIntersection,
    class: classIntersection,
    range: rangeIntersection,
    narrow: narrowIntersection
}, {
    onEmpty: "bubble"
});
const ruleFlatteners = {
    objectKind: flattenObjectKind,
    regex: (entries, rule)=>{
        for (const source of listFrom(rule)){
            entries.push([
                "regex",
                getRegex(source)
            ]);
        }
    },
    divisor: (entries, rule)=>{
        entries.push([
            "divisor",
            rule
        ]);
    },
    range: flattenRange,
    class: (entries, rule)=>{
        entries.push([
            "class",
            rule
        ]);
    },
    props: flattenProps,
    narrow: (entries, rule)=>{
        for (const narrow of listFrom(rule)){
            entries.push([
                "narrow",
                narrow
            ]);
        }
    },
    value: (entries, rule)=>{
        entries.push([
            "value",
            rule
        ]);
    }
};
export const precedenceMap = {
    // Critical: No other checks are performed if these fail
    domain: 0,
    value: 0,
    domains: 0,
    branches: 0,
    objectKind: 0,
    switch: 0,
    alias: 0,
    // Shallow: All shallow checks will be performed even if one or more fail
    class: 1,
    regex: 1,
    divisor: 1,
    bound: 1,
    // Deep: Performed if all shallow checks pass, even if one or more deep checks fail
    requiredProps: 2,
    optionalProps: 2,
    // Narrow: Only performed if all shallow and deep checks pass
    narrow: 3,
    // Morph: Only performed if all validation passes
    morph: 4
};
export const flattenBranch = (branch, ctx)=>{
    if ("morph" in branch) {
        const result = flattenRules(branch.input, ctx);
        result.push([
            "morph",
            branch.morph
        ]);
        return result;
    }
    return flattenRules(branch, ctx);
};
const flattenRules = (rules, ctx)=>{
    const entries = [];
    let k;
    for(k in rules){
        ruleFlatteners[k](entries, rules[k], ctx);
    }
    return entries.sort((l, r)=>precedenceMap[l[0]] - precedenceMap[r[0]]);
};
export const literalSatisfiesRules = (data, rules, state)=>!state.type.meta.scope.type([
        "node",
        {
            [state.domain]: rules
        }
    ])(data).problems;
