import type { ResolvedNode, TraversalNode, TypeNode } from "./nodes/node.js";
import type { as, inferDefinition, ParseContext, validateDefinition } from "./parse/definition.js";
import type { ParsedMorph } from "./parse/tuple/morph.js";
import type { Problems, ProblemsConfig, ProblemsOptions } from "./traverse/problems.js";
import type { defer, Dict, error, evaluateObject, isAny, List, nominal } from "./utils/generics.js";
import type { BuiltinClass, InferredObjectKinds } from "./utils/objectKinds.js";
import type { stringifyUnion } from "./utils/unionToTuple.js";
type ScopeParser = {
    <aliases>(aliases: validateAliases<aliases, {}>): Scope<parseScope<aliases, {}>>;
    <aliases, opts extends ScopeOptions>(aliases: validateAliases<aliases, opts>, opts: validateOptions<opts>): Scope<parseScope<aliases, opts>>;
};
export type TypeParser<$> = {
    <def>(def: validateDefinition<def, $>): parseType<def, $>;
    <def>(def: validateDefinition<def, $>, opts: TypeOptions): parseType<def, $>;
};
export type ScopeOptions = {
    imports?: Space[] | [];
    includes?: Space[] | [];
    standard?: boolean;
    name?: string;
};
type validateOptions<opts extends ScopeOptions> = {
    [k in keyof opts]: k extends "imports" | "includes" ? mergeSpaces<opts[k]> extends error<infer e> ? e : opts[k] : opts[k];
};
export type ScopeContext = Dict | ScopeContextTuple;
type ScopeContextTuple = [exports: Dict, locals: Dict, standard?: false];
type parseScope<aliases, opts extends ScopeOptions> = opts["standard"] extends false ? [inferExports<aliases, opts>, importsOf<opts>, false] : opts["imports"] extends Space[] ? [inferExports<aliases, opts>, importsOf<opts>] : inferExports<aliases, opts>;
type importsOf<opts extends ScopeOptions> = unknown extends opts["imports"] ? {} : mergeSpaces<opts["imports"]>;
type includesOf<opts extends ScopeOptions> = unknown extends opts["includes"] ? {} : mergeSpaces<opts["includes"]>;
export type resolve<name extends keyof $, $> = isAny<$[name]> extends true ? any : $[name] extends alias<infer def> ? inferDefinition<def, $> : $[name];
type exportsOf<context extends ScopeContext> = context extends [
    infer exports,
    ...unknown[]
] ? exports : context;
type localsOf<context extends ScopeContext> = context extends List ? context["1"] & (context["2"] extends false ? {} : PrecompiledDefaults) : PrecompiledDefaults;
type mergeSpaces<scopes, base extends Dict = {}> = scopes extends readonly [
    Space<infer head>,
    ...infer tail
] ? keyof base & keyof head extends never ? mergeSpaces<tail, base & head> : error<`Duplicates ${stringifyUnion<keyof base & keyof head & string>}`> : base;
type validateAliases<aliases, opts extends ScopeOptions> = {
    [name in keyof aliases]: name extends keyof preresolved<opts> ? writeDuplicateAliasMessage<name & string> : validateDefinition<aliases[name], bootstrapScope<aliases, opts>>;
};
type preresolved<opts extends ScopeOptions> = includesOf<opts> & importsOf<opts> & (opts["standard"] extends false ? {} : PrecompiledDefaults);
type alias<def = {}> = nominal<def, "alias">;
type bootstrapScope<aliases, opts extends ScopeOptions> = {
    [k in keyof aliases]: alias<aliases[k]>;
} & preresolved<opts>;
type inferExports<aliases, opts extends ScopeOptions> = evaluateObject<{
    [k in keyof aliases]: inferDefinition<aliases[k], bootstrapScope<aliases, opts>>;
} & includesOf<opts>>;
export type Space<exports = Dict> = {
    [k in keyof exports]: Type<exports[k]>;
};
type resolutions<context extends ScopeContext> = localsOf<context> & exportsOf<context>;
type name<context extends ScopeContext> = keyof resolutions<context> & string;
export declare class Scope<context extends ScopeContext = any> {
    #private;
    aliases: Dict;
    opts: ScopeOptions;
    name: string;
    parseCache: FreezingCache<TypeNode>;
    constructor(aliases: Dict, opts: ScopeOptions);
    type: TypeParser<resolutions<context>>;
    createAnonymousTypeName(): string;
    get infer(): exportsOf<context>;
    compile(): Space<exportsOf<context>>;
    addReferenceIfResolvable(name: name<context>, ctx: ParseContext): boolean;
    resolve(name: name<context>): Type<unknown>;
    resolveNode(node: TypeNode): ResolvedNode;
}
type TypeRoot<t = unknown> = {
    [as]: t;
    infer: asOut<t>;
    node: TypeNode;
    flat: TraversalNode;
    meta: TypeMeta;
};
type TypeMeta = {
    name: string;
    id: QualifiedTypeName;
    definition: unknown;
    scope: Scope;
    problems: ProblemsConfig;
    includesMorph: boolean;
};
declare class Cache<item = unknown> {
    protected cache: {
        [name in string]?: item;
    };
    get root(): {
        readonly [name in string]?: item;
    };
    has(name: string): boolean;
    get(name: string): item | undefined;
    set(name: string, item: item): item;
}
declare class FreezingCache<item = unknown> extends Cache<item> {
    set(name: string, item: item): item;
}
export declare const writeShallowCycleErrorMessage: (name: string, seen: string[]) => string;
export declare const scope: ScopeParser;
export declare const scopes: {
    tsKeywords: Scope<[{
        any: any;
        bigint: bigint;
        boolean: boolean;
        false: false;
        never: never;
        null: null;
        number: number;
        object: object;
        string: string;
        symbol: symbol;
        true: true;
        unknown: unknown;
        void: void;
        undefined: undefined;
    }, {}, false]>;
    jsObjects: Scope<[{
        Function: (...args: any[]) => unknown;
        Array: unknown[];
        Date: Date;
        Error: Error;
        Map: Map<unknown, unknown>;
        RegExp: RegExp;
        Set: Set<unknown>;
        Object: {
            [x: string]: unknown;
        };
        String: String;
        Number: Number;
        Boolean: Boolean;
        WeakMap: WeakMap<object, unknown>;
        WeakSet: WeakSet<object>;
        Promise: Promise<unknown>;
    }, {}, false]>;
    validation: Scope<[{
        email: string;
        alphanumeric: string;
        alpha: string;
        lowercase: string;
        uppercase: string;
        integer: number;
    }, {}, false]>;
    standard: Scope<[{
        any: any;
        bigint: bigint;
        boolean: boolean;
        false: false;
        never: never;
        null: null;
        number: number;
        object: object;
        string: string;
        symbol: symbol;
        true: true;
        unknown: unknown;
        void: void;
        undefined: undefined;
        Function: (...args: any[]) => unknown;
        Array: unknown[];
        Date: Date;
        Error: Error;
        Map: Map<unknown, unknown>;
        RegExp: RegExp;
        Set: Set<unknown>;
        Object: {
            [x: string]: unknown;
        };
        String: String;
        Number: Number;
        Boolean: Boolean;
        WeakMap: WeakMap<object, unknown>;
        WeakSet: WeakSet<object>;
        Promise: Promise<unknown>;
        email: string;
        alphanumeric: string;
        alpha: string;
        lowercase: string;
        uppercase: string;
        integer: number;
    }, {}, false]>;
};
export declare const spaces: {
    tsKeywords: Space<{
        any: any;
        bigint: bigint;
        boolean: boolean;
        false: false;
        never: never;
        null: null;
        number: number;
        object: object;
        string: string;
        symbol: symbol;
        true: true;
        unknown: unknown;
        void: void;
        undefined: undefined;
    }>;
    jsObjects: Space<{
        Function: (...args: any[]) => unknown;
        Array: unknown[];
        Date: Date;
        Error: Error;
        Map: Map<unknown, unknown>;
        RegExp: RegExp;
        Set: Set<unknown>;
        Object: {
            [x: string]: unknown;
        };
        String: String;
        Number: Number;
        Boolean: Boolean;
        WeakMap: WeakMap<object, unknown>;
        WeakSet: WeakSet<object>;
        Promise: Promise<unknown>;
    }>;
    validation: Space<{
        email: string;
        alphanumeric: string;
        alpha: string;
        lowercase: string;
        uppercase: string;
        integer: number;
    }>;
    standard: Space<{
        any: any;
        bigint: bigint;
        boolean: boolean;
        false: false;
        never: never;
        null: null;
        number: number;
        object: object;
        string: string;
        symbol: symbol;
        true: true;
        unknown: unknown;
        void: void;
        undefined: undefined;
        Function: (...args: any[]) => unknown;
        Array: unknown[];
        Date: Date;
        Error: Error;
        Map: Map<unknown, unknown>;
        RegExp: RegExp;
        Set: Set<unknown>;
        Object: {
            [x: string]: unknown;
        };
        String: String;
        Number: Number;
        Boolean: Boolean;
        WeakMap: WeakMap<object, unknown>;
        WeakSet: WeakSet<object>;
        Promise: Promise<unknown>;
        email: string;
        alphanumeric: string;
        alpha: string;
        lowercase: string;
        uppercase: string;
        integer: number;
    }>;
};
export type PrecompiledDefaults = {
    email: string;
    alphanumeric: string;
    alpha: string;
    lowercase: string;
    uppercase: string;
    integer: number;
    any: any;
    bigint: bigint;
    boolean: boolean;
    false: false;
    never: never;
    null: null;
    number: number;
    object: object;
    string: string;
    symbol: symbol;
    true: true;
    unknown: unknown;
    void: void;
    undefined: undefined;
} & InferredObjectKinds;
export declare const type: TypeParser<PrecompiledDefaults>;
export declare const writeDuplicateAliasMessage: <name_1 extends string>(name: name_1) => `Alias '${name_1}' is already defined`;
type writeDuplicateAliasMessage<name extends string> = `Alias '${name}' is already defined`;
export declare const isType: (value: unknown) => value is Type<unknown>;
export type parseType<def, $> = def extends validateDefinition<def, $> ? Type<inferDefinition<def, $>> : never;
export type CheckResult<t> = ValidCheckResult<t> | InvalidCheckResult;
type ValidCheckResult<t> = {
    data: asIn<t>;
    out: asOut<t>;
    problems?: never;
};
type InvalidCheckResult = {
    data: unknown;
    problems: Problems;
    out?: never;
};
type Checker<t> = (data: unknown) => CheckResult<t>;
export type Type<t = unknown> = defer<Checker<t> & TypeRoot<t>>;
export type TypeOptions = {
    name?: string;
    problems?: ProblemsOptions;
};
export type QualifiedTypeName = `${string}.${string}`;
export type TypeConfig = {
    name: string;
    id: QualifiedTypeName;
    problems: ProblemsConfig;
};
export type asIn<t> = asIo<t, "in">;
export type asOut<t> = asIo<t, "out">;
type asIo<t, io extends "in" | "out"> = t extends ParsedMorph<infer i, infer o> ? io extends "in" ? i : o : t extends object ? t extends Function | BuiltinClass ? t : {
    [k in keyof t]: asIo<t[k], io>;
} : t;
export {};
